"""distutils.command

Package containing implementation of all the standard Distutils
commands."""

__all__ = [  # noqa: F822
    'build',
    'build_py',
    'build_ext',
    'build_clib',
    'build_scripts',
    'clean',
    'install',
    'install_lib',
    'install_headers',
    'install_scripts',
    'install_data',
    'sdist',
    'register',
    'bdist',
    'bdist_dumb',
    'bdist_rpm',
    'bdist_wininst',
    'check',
    'upload',
]
